#ifndef SOLVER_HPP_
#define SOLVER_HPP_


#include "array_set.hpp"
#include "array_map.hpp"


namespace ics {


//Given a solution (a mapping of aspects to their values), return the solution
//  as a nicely formatted std::string (via assignment_to).
template<class Aspect, class Value>
std::string illustrate_solution (ics::ArrayMap<Aspect,Value> assignments) {
  if (assignments.empty())
    return "No solution found";

  std::string solution = "Bindings:";
  for (auto& kv : assignments)
    solution += "\n  " + kv.first.assignment_to(kv.second);

  return solution;
}



//Find the first solution to a problem and return it.
//The body of solve_it and solveAll are very similar.
template<class Aspect, class Value>
ics::ArrayMap<Aspect,Value> solve_it (ics::Problem problem) {
  try {
    //Choose an aspect (and exclude it from being chosen again)
    Aspect aspect = problem.choose_and_exclude_aspect();

    //For every possible value for that aspect, recursively attempt to
    //  solve a copy of the problem with that aspect bound to that value
    //If a solution is found, return it; if not, unbind the aspect and
    //  continue looping (finding a new value to bind to the aspect).
    //If Succeed or Fail is thrown, return the answer or null appropriately)
    for (auto v : aspect.choices()) {
      problem.assign_aspect_value(aspect, v);
      ics::ArrayMap<Aspect,Value> temp = solve_it<Aspect,Value>(problem.copy());
      if (temp.size() != 0)
        return temp;
      problem.remove_aspect_value(aspect);
    }
  } catch (ics::BacktrackSucceed& s){return problem.get_solution();}
    catch (ics::BacktrackFail&    f){return ics::ArrayMap<Aspect,Value>();}
  return ics::ArrayMap<Aspect,Value>();
}



//Find all solutions to a problem and return a set of them.
//The body of solve_it and solve_all are very similar
template<class Aspect, class Value>
void solve_all (ics::Problem problem, ics::ArraySet<ics::ArrayMap<Aspect,Value>>& solutions) {
  try {
    //Choose an aspect (and exclude it from being chosen again)
   Aspect aspect = problem.choose_and_exclude_aspect();

   //For every possible value for that aspect, recursively attempt to
   //  solve a copy of the problem with that aspect bound to that value
   //If a solution is found, return it; if not, unbind the aspect and
   //  continue looping (finding a new value to bind to the aspect).
   //If Succeed or Fail is thrown, return the answer or null appropriately)
   for (auto v : aspect.choices()) {
      problem.assign_aspect_value(aspect, v);
      solve_all<Aspect,Value>(problem.copy(), solutions);
      problem.remove_aspect_value(aspect);
    }
  } catch (ics::BacktrackSucceed& s){solutions.insert(problem.get_solution());}
    catch (ics::BacktrackFail&    f){}
}

}

#endif /* SOLVER_HPP_ */
